# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute exponentially weighted variance."""
from typing import Optional

import haiku as hk
import jax.numpy as jnp


class EWMVar(hk.Module):
    """Compute exponentially weighted variance.

    To calculate the variance we use the fact that Var(X) = Mean(x^2) - Mean(x)^2 and internally
    we use the exponentially weighted mean of x/x^2 to calculate this.

    References
    ----------
    [^1]: [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf) # noqa
    """

    def __init__(
        self,
        *,
        com: Optional[float] = None,
        alpha: Optional[float] = None,
        adjust: bool = True,
        name: Optional[str] = None
    ):
        r"""Initialize the module.

        Args:
            com : Specify decay in terms of center of mass
                :math:`\alpha = 1 / (1 + com)`, for :math:`com \geq 0`.

            alpha:  Specify smoothing factor :math:`\alpha` directly
                :math:`0 < \alpha \leq 1`.

            adjust: if true, implement a non-stationary filter with exponential initialization
                scheme. If "linear", implement a non-stationary filter with linear initialization.

            name : name of the module instance.
        """
        super().__init__(name=name)
        assert (
            com is not None or alpha is not None
        ), "com or alpha parameters must be specified."
        if com is not None:
            assert alpha is None
        elif alpha is not None:
            assert com is None
            com = 1.0 / alpha - 1.0

        self.com = com
        self.adjust = adjust

    def __call__(self, x):
        """Implement last formula in [1]:
        .. maths::
            diff := x - mean
            incr := alpha * diff
            mean := mean + incr
            variance := (1 - alpha) * (variance + diff * incr)
        """

        logcom = hk.get_parameter(
            "logcom",
            shape=[],
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.array(jnp.log(self.com), dtype),
        )
        com = jnp.exp(logcom)
        alpha = 1.0 / (1.0 + com)

        mean = hk.get_state(
            "mean",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, jnp.nan, dtype),
        )
        variance = hk.get_state(
            "variance",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, jnp.nan, dtype),
        )
        nobs = hk.get_state(
            "nobs",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, 0.0, dtype),
        )

        # initialization on first non-nan value
        mean = jnp.where(jnp.isnan(mean), x, mean)
        variance = jnp.where(jnp.isnan(variance), 0.0, variance)

        mask = jnp.logical_not(jnp.isnan(x))
        nobs = jnp.where(mask, nobs + 1, nobs)

        # alpha adjustement scheme
        if self.adjust == "linear":
            tscale = 1.0 / alpha
            tscale = jnp.where(nobs < tscale, nobs, tscale)
            alpha = jnp.where(tscale > 0, 1.0 / tscale, jnp.nan)
        elif self.adjust:
            # exponential scheme (as in pandas)
            alpha = alpha / (1.0 - (1.0 - alpha) ** (nobs))

        diff = x - mean
        incr = alpha * diff

        # update state
        mean = jnp.where(mask, mean + incr, mean)
        variance = jnp.where(mask, (1 - alpha) * (variance + diff * incr), variance)

        hk.set_state("mean", mean)
        hk.set_state("variance", variance)
        hk.set_state("nobs", nobs)

        return variance



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Normalize data by its standard deviation computed with a mask."""
import haiku as hk
from jax import tree_map

from wax.modules.fill_nan_inf import FillNanInf
from wax.modules.mask_std import MaskStd


class MaskNormalize(hk.Module):
    """Normalize data by its standard deviation computed with a mask."""

    def __init__(self, axis=None, assume_centered=False, name=None):
        """Initialize the module.

        Args:
            axis : axis along which to compute the mean
            assume_centered : if True assume mean to be zero when computing
                the standard deviation.
        """
        super().__init__(name=name)
        self.axis = axis
        self.assume_centered = assume_centered

    def __call__(self, mask, input):
        """Normalize data.

        Args:
            mask : bolean mask specifying points over which to compute the mean.
            input : data on which to compute the mean.
        """

        def normalize(x):
            x_std = MaskStd(axis=self.axis, assume_centered=self.assume_centered)(
                mask, x
            )
            x_normalized = FillNanInf()(x / x_std)
            return x_normalized

        return tree_map(normalize, input)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transformation functions to work on batches of data."""

from typing import Any, Callable, NamedTuple

import jax
from jax import tree_map

from wax.unroll import static_scan


class TransformedBatch(NamedTuple):
    init: Callable
    apply: Callable


class BatchState(NamedTuple):
    fun_params: Any
    fun_state: Any
    rng_key: Any


class BatchParams(NamedTuple):
    ...


def transform_batch_with_state(fun, skip_first=False):
    """Transforms a pair of pure functions into another pair of pure functions
    which run on a batch.
    """

    def scan_f(prev_state, inputs):
        fun_params, fun_state, key = prev_state
        outputs, fun_state = fun.apply(fun_params, fun_state, key, inputs)
        return BatchState(fun_params, fun_state, key), outputs

    def init(key, xs):
        x0 = tree_map(lambda x: x[0], xs)
        fun_params, fun_state = fun.init(key, x0)
        return BatchParams(), BatchState(fun_params, fun_state, key)

    def apply(params, state, key, xs):
        # params is not used ... (and is empty)
        # replace state rng_key by given rng_key
        # we drop the first observation because it has been used in initialization....
        if skip_first:
            xs = tree_map(lambda x: x[1:], xs)
        final_state, output_sequence = jax.lax.scan(scan_f, init=state, xs=xs)
        assert isinstance(final_state, BatchState)
        return output_sequence, final_state

    return TransformedBatch(init, apply)


def transform_batch_with_state_static(fun, skip_first=False, pbar=True):
    """Transforms a pair of pure functions into another pair of pure functions
    which run on a batch.
    """

    def scan_f(prev_state, inputs):
        fun_params, fun_state, key = prev_state
        outputs, fun_state = fun.apply(fun_params, fun_state, key, inputs)
        return BatchState(fun_params, fun_state, key), outputs

    def init(key, xs):
        x0 = tree_map(lambda x: x[0], xs)
        fun_params, fun_state = fun.init(key, x0)
        return BatchParams(), BatchState(fun_params, fun_state, key)

    def apply(params, state, key, xs):
        # params is not used ... (and is empty)
        # replace state rng_key by given rng_key
        init_state = BatchState(state.fun_params, state.fun_state, key)
        # we drop the first observation because it has been used in initialization....
        if skip_first:
            xs = tree_map(lambda x: x[1:], xs)
        final_state, output_sequence = static_scan(scan_f, init=init_state, xs=xs)
        return output_sequence, final_state

    return TransformedBatch(init, apply)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import pytest
from jax import numpy as jnp

from wax.compile import jit_init_apply
from wax.modules.lag import Lag
from wax.unroll import unroll


@pytest.mark.parametrize("use_jit", [False, True])
def test_lag(use_jit):
    seq = hk.PRNGSequence(42)
    x = jax.random.normal(next(seq), (2, 3))

    @hk.transform_with_state
    def lag(x):
        return Lag(1)(x)

    if use_jit:
        lag = jit_init_apply(lag)
    params, state = lag.init(next(seq), x)
    output, state = lag.apply(params, state, next(seq), x)
    assert len(output) == 2
    assert jnp.isnan(output).all()

    x1 = x
    x = jax.random.normal(next(seq), (2, 3))
    output, state = lag.apply(params, state, next(seq), x)
    assert len(output) == 2
    assert (output == (x1)).all()


def test_lag_unroll_int():
    xs = jnp.array([10, 11, 12], dtype="int32")

    res = unroll(lambda x: Lag(1)(x))(xs)
    assert (res == jnp.array([0, 10, 11])).all()

    res = unroll(lambda x: Lag(2)(x))(xs)
    assert (res == jnp.array([0, 0, 10])).all()


def test_lag_unroll_float():
    xs = jnp.array([10, 11, 12], dtype="float32")

    res = unroll(lambda x: Lag(1)(x))(xs)
    assert jnp.allclose(res, jnp.array([jnp.nan, 10, 11]), equal_nan=True)

    res = unroll(lambda x: Lag(2)(x))(xs)
    assert jnp.allclose(res, jnp.array([jnp.nan, jnp.nan, 10]), equal_nan=True)

    res = unroll(lambda x: Lag(1)(Lag(1)(x)))(xs)
    assert jnp.allclose(res, jnp.array([jnp.nan, jnp.nan, 10]), equal_nan=True)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define API for Gym agents."""
import logging
from abc import ABCMeta, abstractmethod
from typing import Any

from wax.gym.entity import Entity
from wax.gym.gym_unroll import gym_unroll

logger = logging.getLogger(__name__)


# @dataclass
class Agent(Entity, metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, obs: Any) -> Any:
        ...

    def reset(self):
        logger.warning(f"`reset` for agent `{self.name}` not implemented")

    def feed(self, reward):
        logger.warning(f"`feed` for agent `{self.name}` not implemented")

    def unroll(
        self,
        env,
        *,
        obs=None,
        callbacks=None,
    ):
        """Unroll agent with a gym environment.

        Parameters
        ----------
        env
            Gym environment used to unroll the data and feed the agent.

        obs
            if passed, start the Gym unroll loop with this observation instead of
            the observation returned by 'env.reset()', else 'obs = env.reset()' is
            used as first observation.

        callbacks
            Additional callback to use


        Returns
        -------
        gym_state
            nested data structure
        """
        return gym_unroll(self, env, obs=obs, callbacks=callbacks)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implement buffering mechanism."""

from typing import Any, Callable, NamedTuple, Optional

import haiku as hk
import jax.numpy as jnp


class BufferState(NamedTuple):
    buffer: Any
    len_buffer: int
    i_start: int


class BufferFun(NamedTuple):
    init: Callable
    apply: Callable


def buffer_fn(maxlen: int, fill_value=jnp.nan):
    def init(shape, dtype):
        buffer = jnp.full((maxlen,) + shape, fill_value, dtype=dtype)
        len_buffer = 0
        i_start = maxlen
        return BufferState(buffer, len_buffer, i_start)

    def apply(x, state):
        buffer, len_buffer, i_start = state

        buffer = jnp.roll(buffer, -1, axis=0)
        buffer = buffer.at[-1].set(x)
        len_buffer = jnp.minimum(len_buffer + 1, maxlen)
        i_start = maxlen - len_buffer

        return buffer, BufferState(buffer, len_buffer, i_start)

    return BufferFun(init, apply)


class Buffer(hk.Module):
    """Implement buffering mechanism."""

    def __init__(
        self,
        maxlen: int,
        fill_value=jnp.nan,
        return_state: bool = False,
        name: Optional[str] = None,
    ):
        """Initialize the module/

        Args:
            maxlen : length of the buffer
            fill_value : value to use to fill buffer while no data has been append.
            return_state : if true, the module returns a tuple (buffer, state) where state is
                           the full buffer state (buffer, len_buffer, i_start). If false, the buffer is
                           returned.
            name : name of the module.
        """
        super().__init__(name=name)
        self.maxlen = maxlen
        self.fill_value = fill_value
        self.return_state = return_state

    def __call__(self, input: jnp.ndarray):
        """Record input data in the buffer.

        Args:
            input: data to record.
        """

        fun = buffer_fn(self.maxlen, self.fill_value)

        buffer_state = hk.get_state(
            "buffer_state",
            input.shape,
            input.dtype,
            init=fun.init,
        )

        buffer, buffer_state = fun.apply(input, buffer_state)

        hk.set_state("buffer_state", buffer_state)
        if self.return_state:
            return buffer, buffer_state
        else:
            return buffer



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Apply mask"""
from functools import partial
from typing import Any

import haiku as hk
import jax.numpy as jnp
from jax import tree_map


class ApplyMask(hk.Module):
    """Apply mask"""

    def __init__(self, axis=None, mask_value=0.0, name=None):
        """Initialize module.

        Args:
            axis: if not None, axis along which applying the mask.
            mask_value: value to return where mask is false.
            name : name of the module instance.
        """
        super().__init__(name=name)
        self.axis = axis
        self.mask_value = mask_value

    def __call__(self, mask: jnp.ndarray, input: Any):
        """Apply mask.

        Args:
            mask: mask data array.
            input: nested data structure on which we apply the mask
        """

        def apply_mask(mask, x):
            # mask values
            if self.axis is not None:
                if self.axis == 0:
                    mask = jnp.repeat(mask.reshape(-1, 1), x.shape[1], axis=1)
                elif self.axis == 1:
                    mask = jnp.repeat(mask.reshape(1, -1), x.shape[0], axis=0)
                else:
                    # TODO: implement for any value of axis
                    raise ValueError(
                        "ApplyMask is not implemented for "
                        "axis different from None, 0, 1."
                    )

            if x.shape != mask.shape:
                raise ValueError("mask shape and data shape are not coherent")
            return jnp.where(mask, x, self.mask_value)

        return tree_map(partial(apply_mask, mask), input)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
from jax.numpy import array as DeviceArray
from jax.numpy import uint32


class Counter(hk.Module):
    def __call__(self):
        count = hk.get_state(
            "count", [], init=lambda shape, dtype: DeviceArray(0, dtype=uint32)
        )
        count += 1
        hk.set_state("count", count)
        return count



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines a comparison function to compare two nested data structures."""
from numbers import Number

import numpy as onp
from jax import numpy as jnp
from jax._src.numpy.lax_numpy import int32, uint32
from jax.tree_util import tree_flatten


def assert_tree_all_close(x, y, check_treedef=True):
    from jax.config import config

    config.update("jax_enable_x64", False)

    x, x_treedef = tree_flatten(x)
    y, y_treedef = tree_flatten(y)
    if check_treedef:
        ref_str = str(x_treedef).replace("wax.", "")
        check_str = str(y_treedef).replace("_test", "").replace("wax.", "")
        assert ref_str == check_str, f"{ref_str} != {check_str}"
    assert len(x) == len(y)
    for x_, y_ in zip(x, y):
        assert type(x_) == type(y_)

        if isinstance(x_, Number):
            assert x_ == y_
        else:
            assert x_.dtype == y_.dtype, f"{x_.dtype} != {y_.dtype}"
            if x_.dtype in [int32, uint32]:
                assert (x_ == y_).all()
            elif isinstance(x, jnp.ndarray):
                assert jnp.allclose(x_, y_)
            else:
                assert onp.allclose(x_, y_)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute exponentially weighted covariance."""
from typing import Optional

import haiku as hk
import jax.numpy as jnp

from wax.modules.ewma import EWMA


class EWMCov(hk.Module):
    """Compute exponentially weighted covariance.

    To calculate the variance we use the fact that Var(X) = Mean(x^2) - Mean(x)^2 and internally
    we use the exponentially weighted mean of x/x^2 to calculate this.

    References
    ----------
    [^1]: [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf) # noqa
    """

    def __init__(
        self,
        *,
        com: Optional[float] = None,
        alpha: Optional[float] = None,
        min_periods: int = 0,
        adjust: bool = True,
        ignore_na: bool = False,
        initial_value: float = jnp.nan,
        assume_centered: bool = False,
        name: Optional[str] = None,
    ):
        r"""Initialize the module.

        Args:
            com : Specify decay in terms of center of mass
                :math:`\alpha = 1 / (1 + com)`, for :math:`com \geq 0`.

            alpha:  Specify smoothing factor :math:`\alpha` directly
                :math:`0 < \alpha \leq 1`.

            min_periods : Minimum number of observations in window required to have a value;
                otherwise, result is ``np.nan``.

            adjust : Divide by decaying adjustment factor in beginning periods to account
                for imbalance in relative weightings (viewing EWMA as a moving average).
                - When ``adjust=True`` (default), the EW function is calculated using weights
                  :math:`w_i = (1 - \alpha)^i`. For example, the EW moving average of the series
                  [:math:`x_0, x_1, ..., x_t`] would be:
                .. math::
                    y_t = \frac{x_t + (1 - \alpha)x_{t-1} + (1 - \alpha)^2 x_{t-2} + ... + (1 -
                    \alpha)^t x_0}{1 + (1 - \alpha) + (1 - \alpha)^2 + ... + (1 - \alpha)^t}
                - When ``adjust=False``, the exponentially weighted function is calculated
                  recursively:
                .. math::
                    \begin{split}
                        y_0 &= x_0\\
                        y_t &= (1 - \alpha) y_{t-1} + \alpha x_t,
                    \end{split}
                The effective  center of mass (com) interpolate exponentially between 0 and the
                nominal center of mass.

                - When ``adjust='linear'`` the effective  center of mass (com) interpolate linearly
                between 0 and the nominal center of mass.

            ignore_na : Ignore missing values when calculating weights.
                - When ``ignore_na=False`` (default), weights are based on absolute positions.
                  For example, the weights of :math:`x_0` and :math:`x_2` used in calculating
                  the final weighted average of [:math:`x_0`, None, :math:`x_2`] are
                  :math:`(1-\alpha)^2` and :math:`1` if ``adjust=True``, and
                  :math:`(1-\alpha)^2` and :math:`\alpha` if ``adjust=False``.
                - When ``ignore_na=True``, weights are based
                  on relative positions. For example, the weights of :math:`x_0` and :math:`x_2`
                  used in calculating the final weighted average of
                  [:math:`x_0`, None, :math:`x_2`] are :math:`1-\alpha` and :math:`1` if
                  ``adjust=True``, and :math:`1-\alpha` and :math:`\alpha` if ``adjust=False``.


            initial_value : initial value for the state.

            assume_centered: if true, assume that the mean estimator is zero.

            name : name of the module instance.
        """
        super().__init__(name=name)
        assert (
            com is not None or alpha is not None
        ), "com or alpha parameters must be specified."
        if com is not None:
            assert alpha is None
        elif alpha is not None:
            assert com is None
            com = 1.0 / alpha - 1.0

        self.com = com
        self.min_periods = min_periods
        self.adjust = adjust
        self.ignore_na = ignore_na
        self.initial_value = initial_value
        self.assume_centered = assume_centered

    def __call__(self, x, y=None):
        """Compute"""
        if y is None:
            import warnings

            warnings.warn(
                "x and y arguments may be passed directly"
                "instead as a tuple argument. The tuple argument syntax"
                "may be removed in the future."
            )
            x, y = x
        mean_xy = EWMA(
            com=self.com,
            min_periods=self.min_periods,
            adjust=self.adjust,
            ignore_na=self.ignore_na,
            initial_value=self.initial_value,
            name="mean_xy",
        )(jnp.outer(x, y))
        if self.assume_centered:
            cov = mean_xy
        else:
            mean_x = EWMA(
                com=self.com,
                min_periods=self.min_periods,
                adjust=self.adjust,
                ignore_na=self.ignore_na,
                initial_value=self.initial_value,
                name="mean_x",
            )(x)
            mean_y = EWMA(
                com=self.com,
                min_periods=self.min_periods,
                adjust=self.adjust,
                ignore_na=self.ignore_na,
                initial_value=self.initial_value,
                name="mean_y",
            )(y)
            cov = mean_xy - jnp.outer(mean_x, mean_y)
        return cov



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp

from wax.unroll import unroll


class GenModel(hk.Module):
    def __call__(self, i):
        return hk.next_rng_key()[0]


def test_unroll_rng():
    fun = hk.transform(lambda x: (GenModel()(x), GenModel()(x)))
    rng = jax.random.PRNGKey(42)
    params = fun.init(rng, 1)
    ref = []
    for i in range(5):
        (rng, sub_rng) = jax.random.split(rng)
        ref.append(fun.apply(params, sub_rng, 1))
    ref = jnp.array(ref)

    rng = jax.random.PRNGKey(42)
    out = unroll(lambda x: (GenModel()(x), GenModel()(x)), rng=rng, dynamic=False)(
        jnp.arange(5)
    )
    assert (ref == jnp.stack(out).T).flatten().all()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Online optimizer module."""
from collections import namedtuple
from dataclasses import dataclass
from typing import Any, Callable, NamedTuple, Optional, Union

import haiku as hk
import jax
import jax.numpy as jnp
from haiku.data_structures import partition
from jax.tree_util import tree_map
from optax import GradientTransformation

from wax.modules.optax_optimizer import OptaxOptimizer
from wax.predicate import pass_all_predicate


class ParamsState(NamedTuple):
    trainable_params: Any
    non_trainable_params: Any
    state: Any


@dataclass
class OptInfo:
    """Dynamically define OptimizerInfo namedtuple structure.

    Args:
        return_params: if true, add 'params' field to output structure.
    """

    return_params: bool = False

    def __post_init__(self):
        outputs = ["loss", "model_info"]
        if self.return_params:
            outputs += ["params"]
        self.opt_info_struct_ = namedtuple("OptInfo", outputs)

    def __call__(
        self,
        loss: float,
        model_info: Any,
        params: Any = None,
    ):
        outputs = [loss, model_info]
        if self.return_params:
            outputs += [params]
        return self.opt_info_struct_(*outputs)


class OnlineOptimizer(hk.Module):
    """Wraps a model with loss and and optimizer to perform one online learning update."""

    def __init__(
        self,
        model: Union[Callable, hk.TransformedWithState],
        opt: Union[OptaxOptimizer, GradientTransformation],
        project_params: Optional[Callable] = None,
        regularize_loss: Optional[Callable] = None,
        params_predicate: Callable[[str, str, jnp.ndarray], bool] = pass_all_predicate,
        return_params=False,
        name: Optional[str] = None,
    ):
        """Initialize module.

        Args:
            model : model to optimize. The model should return a tuple (loss, info)
            opt : optimizer: Optax transformation consisting of a function pair: (initialise, update).
            project_params : function to project parameters. It applies to parameters and optimizer state .
            regularize_loss: function to regularize the model loss. It applies to the parameters.
            params_predicate: function to split params in trainable and non-trainable params.
                See https://dm-haiku.readthedocs.io/en/latest/notebooks/non_trainable.html
            name : name of the module
        """
        super().__init__(name=name)
        self.model = (
            model
            if isinstance(model, hk.TransformedWithState)
            else hk.transform_with_state(model)
        )
        self.opt = (
            OptaxOptimizer(opt) if isinstance(opt, GradientTransformation) else opt
        )
        self.project_params = project_params
        self.regularize_loss = regularize_loss
        self.params_predicate = params_predicate

        self.OptInfo = OptInfo(return_params)

    def __call__(self, *args, **kwargs):
        """Update learner.

        Args:
            x : features
            y: target
        """

        def init_model_params_and_state(shape, dtype):
            """Set state from trainable params and state of the model."""
            params, state = self.model.init(hk.next_rng_key(), *args, **kwargs)
            trainable_params, non_trainable_params = partition(
                self.params_predicate, params
            )
            trainable_params = hk.data_structures.to_mutable_dict(trainable_params)
            return ParamsState(trainable_params, non_trainable_params, state)

        trainable_params, non_trainable_params, state = hk.get_state(
            "model_params_and_state",
            [],
            init=init_model_params_and_state,
        )

        step = hk.get_state("step", [], init=lambda *_: 0)

        def _loss(trainable_params, non_trainable_params, state, *args, **kwargs):
            params = hk.data_structures.merge(trainable_params, non_trainable_params)
            (loss, model_info), state = self.model.apply(
                params, state, None, *args, **kwargs
            )
            if self.regularize_loss:
                loss += self.regularize_loss(trainable_params)
            return loss, (model_info, params, state)

        # compute loss and gradients
        (loss, (model_info, params, state)), grads = jax.value_and_grad(
            _loss, has_aux=True
        )(trainable_params, non_trainable_params, state, *args, **kwargs)

        # update optimizer state
        filled_grads = tree_map(jnp.nan_to_num, grads)

        # update params
        updated_trainable_params = self.opt(trainable_params, filled_grads)

        if self.project_params:
            updated_trainable_params = self.project_params(
                updated_trainable_params, self.opt
            )

        updated_params = hk.data_structures.merge(
            updated_trainable_params, non_trainable_params
        )
        step += 1
        hk.set_state("step", step)
        hk.set_state(
            "model_params_and_state",
            ParamsState(updated_trainable_params, non_trainable_params, state),
        )

        opt_info = self.OptInfo(
            loss,
            model_info,
            updated_params,
        )
        return opt_info



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.3
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # 🔄 Online learning in non-stationary environments 🔄
#
# We reproduce the empirical results of [1].
#
# ## References
#
# [1] [Anava, O., Hazan, E., Mannor, S. and Shamir, O., 2013, June. Online learning for time series prediction. In Conference on learning theory (pp. 172-184)](https://arxiv.org/pdf/1302.6927.pdf)
#
#

# %pylab inline
# %load_ext autoreload
# %autoreload 2


# +
from typing import Any, NamedTuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as onp
import optax
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from tqdm.auto import tqdm

from wax.modules import ARMA, SNARIMAX, GymFeedback, OnlineOptimizer, UpdateParams, VMap
from wax.modules.lag import tree_lag
from wax.modules.vmap import add_batch
from wax.optim import newton
from wax.unroll import unroll_transform_with_state

# + tags=["parameters"]
T = 10000
N_BATCH = 20
N_STEP_SIZE = 30
N_STEP_SIZE_NEWTON = 10
N_EPS = 5
# -


# ## Agent

OPTIMIZERS = [optax.sgd, optax.adagrad, optax.rmsprop, optax.adam]


# +
from optax._src.base import OptState


def build_agent(time_series_model=None, opt=None):
    if time_series_model is None:
        time_series_model = lambda y, X: SNARIMAX(10)(y, X)

    if opt is None:
        opt = optax.sgd(1.0e-3)

    class AgentInfo(NamedTuple):
        optim: Any
        forecast: Any

    class ModelWithLossInfo(NamedTuple):
        pred: Any
        loss: Any

    def agent(obs):
        if isinstance(obs, tuple):
            y, X = obs
        else:
            y = obs
            X = None

        def evaluate(y_pred, y):
            return jnp.linalg.norm(y_pred - y) ** 2, {}

        def model_with_loss(y, X=None):
            # predict with lagged data
            y_pred, pred_info = time_series_model(*tree_lag(1)(y, X))

            # evaluate loss with actual data
            loss, loss_info = evaluate(y_pred, y)

            return loss, ModelWithLossInfo(pred_info, loss_info)

        def project_params(params: Any, opt_state: OptState = None):
            del opt_state
            return jax.tree_util.tree_map(lambda w: jnp.clip(w, -1, 1), params)

        def params_predicate(m: str, n: str, p: jnp.ndarray) -> bool:
            # print(m, n, p)
            return m.endswith("snarimax/~/linear") and n == "w"

        def learn_and_forecast(y, X=None):
            # use lagged data for the optimizer
            opt_info = OnlineOptimizer(
                model_with_loss,
                opt,
                project_params=project_params,
                params_predicate=params_predicate,
                return_params=True,
            )(*tree_lag(1)(y, X))

            # use updated params to forecast with actual data
            predict_params = opt_info.params

            y_pred, forecast_info = UpdateParams(time_series_model)(
                predict_params, y, X
            )
            return y_pred, AgentInfo(opt_info, forecast_info)

        return learn_and_forecast(y, X)

    return agent


# -

# ## Non-stationary environments
#

# We will now wrapup the study of an environment + agent in few analysis functions.
#
# We will then use them to perform the same analysis in the non-stationary setting proposed in [1], namely:
#
#   * setting 1 : sanity check (stationary ARMA environment).
#   * setting 2 : slowly varying parameters.
#   * setting 3 : brutal variation of parameters.
#   * setting 4 : non-stationary (random walk) noise.

# ## Analysis functions

# For each solver, we will select the best hyper parameters (step size $\eta$, $\epsilon$)
# by measuring the average loss between the 5000 and 10000 steps.

# ### First order solvers
#
#
#
#
#


def scan_hparams_first_order():

    STEP_SIZE_idx = pd.Index(onp.logspace(-4, 1, N_STEP_SIZE), name="step_size")
    STEP_SIZE = jax.device_put(STEP_SIZE_idx.values)

    rng = jax.random.PRNGKey(42)
    eps = sample_noise(rng)

    res = {}
    for optimizer in tqdm(OPTIMIZERS):

        def gym_loop_scan_hparams(eps):
            def scan_params(step_size):
                return GymFeedback(build_agent(opt=optimizer(step_size)), env)(eps)

            res = VMap(scan_params)(STEP_SIZE)
            return res

        sim = unroll_transform_with_state(add_batch(gym_loop_scan_hparams))

        params, state = sim.init(rng, eps)
        _res, state = sim.apply(params, state, rng, eps)
        res[optimizer.__name__] = _res

    ax = None
    BEST_STEP_SIZE = {}
    BEST_GYM = {}

    for name, (gym, info) in res.items():

        loss = (
            pd.DataFrame(-gym.reward, columns=STEP_SIZE).iloc[LEARN_TIME_SLICE].mean()
        )

        BEST_STEP_SIZE[name] = loss.idxmin()

        best_idx = jnp.argmax(gym.reward[LEARN_TIME_SLICE].mean(axis=0))
        BEST_GYM[name] = jax.tree_util.tree_map(lambda x: x[:, best_idx], gym)

        ax = loss.plot(
            logx=True, logy=False, ax=ax, label=name, ylim=(MIN_ERR, MAX_ERR)
        )
    plt.legend(bbox_to_anchor=(1.0, 1.0))

    return BEST_STEP_SIZE, BEST_GYM


# We will "cross-validate" the result by running the agent on new samples.

CROSS_VAL_RNG = jax.random.PRNGKey(44)

COLORS = sns.color_palette("hls")


def cross_validate_first_order(BEST_STEP_SIZE, BEST_GYM):
    plt.figure()
    eps = sample_noise(CROSS_VAL_RNG)
    CROSS_VAL_GYM = {}
    ax = None

    # def measure(reward):
    #     return pd.Series(-reward).rolling(T/2, min_periods=T/2).mean()

    def measure(reward):
        return pd.Series(-reward).expanding().mean()

    for i, (name, gym) in enumerate(BEST_GYM.items()):
        ax = measure(gym.reward).plot(
            ax=ax,
            color=COLORS[i],
            label=(f"(TRAIN) -  {name}    " f"-    $\eta$={BEST_STEP_SIZE[name]:.2e}"),
            style="--",
        )

    for i, optimizer in enumerate(tqdm(OPTIMIZERS)):

        name = optimizer.__name__

        def gym_loop(eps):
            return GymFeedback(build_agent(opt=optimizer(BEST_STEP_SIZE[name])), env)(
                eps
            )

        sim = unroll_transform_with_state(add_batch(gym_loop))

        rng = jax.random.PRNGKey(42)
        params, state = sim.init(rng, eps)
        (gym, info), state = sim.apply(params, state, rng, eps)
        CROSS_VAL_GYM[name] = gym

        ax = measure(gym.reward).plot(
            ax=ax,
            color=COLORS[i],
            ylim=(MIN_ERR, MAX_ERR),
            label=(
                f"(VALIDATE) -  {name}    " f"-    $\eta$={BEST_STEP_SIZE[name]:.2e}"
            ),
        )
    plt.legend(bbox_to_anchor=(1.0, 1.0))

    return CROSS_VAL_GYM


# ### Newton solver
#
#
#
#
#


def scan_hparams_newton():
    STEP_SIZE = pd.Index(onp.logspace(-2, 3, N_STEP_SIZE_NEWTON), name="step_size")
    EPS = pd.Index(onp.logspace(-4, 3, N_EPS), name="eps")

    HPARAMS_idx = pd.MultiIndex.from_product([STEP_SIZE, EPS])
    HPARAMS = jnp.stack(list(map(onp.array, HPARAMS_idx)))

    @add_batch
    def gym_loop_scan_hparams(eps):
        def scan_params(hparams):
            step_size, newton_eps = hparams
            agent = build_agent(opt=newton(step_size, eps=newton_eps))
            return GymFeedback(agent, env)(eps)

        return VMap(scan_params)(HPARAMS)

    sim = unroll_transform_with_state(gym_loop_scan_hparams)

    rng = jax.random.PRNGKey(42)
    eps = sample_noise(rng)

    params, state = sim.init(rng, eps)
    res_newton, state = sim.apply(params, state, rng, eps)

    gym_newton, info_newton = res_newton

    loss_newton = (
        pd.DataFrame(-gym_newton.reward, columns=HPARAMS_idx)
        .iloc[LEARN_TIME_SLICE]
        .mean()
        .unstack()
    )

    sns.heatmap(loss_newton[loss_newton < 0.4], annot=True, cmap="YlGnBu")

    STEP_SIZE, NEWTON_EPS = loss_newton.stack().idxmin()

    x = -gym_newton.reward[LEARN_TIME_SLICE].mean(axis=0)
    x = jnp.where(jnp.isnan(x), jnp.inf, x)
    I_BEST_PARAM = jnp.argmin(x)

    BEST_NEWTON_GYM = jax.tree_util.tree_map(lambda x: x[:, I_BEST_PARAM], gym_newton)

    print("Best newton parameters: ", STEP_SIZE, NEWTON_EPS)

    return (STEP_SIZE, NEWTON_EPS), BEST_NEWTON_GYM


def cross_validate_newton(BEST_HPARAMS, BEST_NEWTON_GYM):
    (STEP_SIZE, NEWTON_EPS) = BEST_HPARAMS
    plt.figure()

    # def measure(reward):
    #     return pd.Series(-reward).rolling(T/2, min_periods=T/2).mean()

    def measure(reward):
        return pd.Series(-reward).expanding().mean()

    @add_batch
    def gym_loop(eps):
        agent = build_agent(opt=newton(STEP_SIZE, eps=NEWTON_EPS))
        return GymFeedback(agent, env)(eps)

    sim = unroll_transform_with_state(gym_loop)

    rng = jax.random.PRNGKey(44)
    eps = sample_noise(rng)
    params, state = sim.init(rng, eps)
    (gym, info), state = sim.apply(params, state, rng, eps)

    ax = None
    i = 4
    ax = measure(BEST_NEWTON_GYM.reward).plot(
        ax=ax,
        color=COLORS[i],
        label=f"(TRAIN) -  Newton    -    $\eta$={STEP_SIZE:.2e},    $\epsilon$={NEWTON_EPS:.2e}",
        ylim=(MIN_ERR, MAX_ERR),
        style="--",
    )

    ax = measure(gym.reward).plot(
        ax=ax,
        color=COLORS[i],
        ylim=(MIN_ERR, MAX_ERR),
        label=f"(VALIDATE) - Newton    -    $\eta$={STEP_SIZE:.2e},    $\epsilon$={NEWTON_EPS:.2e}",
    )
    ax.legend(bbox_to_anchor=(1.0, 1.0))
    ax.plot()
    return gym


# ### Plot everithing
#
#
#
#
#


def plot_everything(BEST_STEP_SIZE, BEST_GYM, BEST_HPARAMS, BEST_NEWTON_GYM):
    MESURES = []

    def measure(reward):
        return pd.Series(-reward).rolling(int(T / 2), min_periods=int(T / 2)).mean()

    MESURES.append(("Rolling mean of loss (5000) time-steps", measure))

    def measure(reward):
        return pd.Series(-reward).expanding().mean()

    MESURES.append(("Expanding means", measure))

    for MEASURE_NAME, MEASUR_FUNC in MESURES:

        plt.figure()

        for i, (name, gym) in enumerate(BEST_GYM.items()):
            MEASUR_FUNC(gym.reward).plot(
                label=f"{name}    -    $\eta$={BEST_STEP_SIZE[name]:.2e}",
                ylim=(MIN_ERR, MAX_ERR),
                color=COLORS[i],
            )

        i = 4
        (STEP_SIZE, NEWTON_EPS) = BEST_HPARAMS
        gym = BEST_NEWTON_GYM
        MEASUR_FUNC(gym.reward).plot(
            label=f"Newton    -    $\eta$={STEP_SIZE:.2e},    $\epsilon$={NEWTON_EPS:.2e}",
            ylim=(MIN_ERR, MAX_ERR),
            color=COLORS[i],
        )
        plt.legend(bbox_to_anchor=(1.0, 1.0))
        plt.title(MEASURE_NAME)


# ## Setting 1

# ### Environment

# let's wrapup the results for the "setting 1" in [1]

# +
from wax.modules import Counter


def build_env():
    def env(action, obs):
        y_pred, eps = action, obs

        ar_coefs = jnp.array([0.6, -0.5, 0.4, -0.4, 0.3])
        ma_coefs = jnp.array([0.3, -0.2])

        y = ARMA(ar_coefs, ma_coefs)(eps)

        rw = -((y - y_pred) ** 2)

        env_info = {"y": y, "y_pred": y_pred}
        obs = y
        return rw, obs, env_info

    return env


def sample_noise(rng):
    eps = jax.random.normal(rng, (T, 20)) * 0.3
    return eps


MIN_ERR = 0.09
MAX_ERR = 0.15
LEARN_TIME_SLICE = slice(int(T / 2), T)
env = build_env()
# -

BEST_STEP_SIZE, BEST_GYM = scan_hparams_first_order()

CROSS_VAL_GYM = cross_validate_first_order(BEST_STEP_SIZE, BEST_GYM)

BEST_HPARAMS, BEST_NEWTON_GYM = scan_hparams_newton()

CROSS_VAL_GYM = cross_validate_newton(BEST_HPARAMS, BEST_NEWTON_GYM)

plot_everything(BEST_STEP_SIZE, BEST_GYM, BEST_HPARAMS, BEST_NEWTON_GYM)


# ### Conclusions
#
# - The NEWTON and ADAGRAD optimizers are the faster to converge.
# - The SGD and ADAM optimizers have the worst performance.

# ### Fixed setting

# +
@add_batch
def gym_loop_newton(eps):
    return GymFeedback(build_agent(opt=newton(0.1, eps=0.3)), env)(eps)


def run_fixed_setting():
    rng = jax.random.PRNGKey(42)
    eps = sample_noise(rng)
    sim = unroll_transform_with_state(gym_loop_newton)
    params, state = sim.init(rng, eps)
    (gym, info), state = sim.apply(params, state, rng, eps)

    pd.Series(-gym.reward).expanding().mean().plot()  # ylim=(MIN_ERR, MAX_ERR))


# -


# %%time
run_fixed_setting()


# ## Setting 2

# ### Environment

# let's build an environment corresponding to "setting 2" in [1]

# +
from wax.modules import Counter


def build_env():
    def env(action, obs):
        y_pred, eps = action, obs
        t = Counter()()
        ar_coefs_1 = jnp.array([-0.4, -0.5, 0.4, 0.4, 0.1])
        ar_coefs_2 = jnp.array([0.6, -0.4, 0.4, -0.5, 0.5])
        ar_coefs = ar_coefs_1 * t / T + ar_coefs_2 * (1 - t / T)

        ma_coefs = jnp.array([0.32, -0.2])

        y = ARMA(ar_coefs, ma_coefs)(eps)
        # prediction used on a fresh y observation.
        rw = -((y - y_pred) ** 2)

        env_info = {"y": y, "y_pred": y_pred}
        obs = y
        return rw, obs, env_info

    return env


def sample_noise(rng):
    eps = jax.random.uniform(rng, (T, 20), minval=-0.5, maxval=0.5)
    return eps


MIN_ERR = 0.0833
MAX_ERR = 0.15
LEARN_TIME_SLICE = slice(int(T / 2), T)
env = build_env()
# -

BEST_STEP_SIZE, BEST_GYM = scan_hparams_first_order()

CROSS_VAL_GYM = cross_validate_first_order(BEST_STEP_SIZE, BEST_GYM)

BEST_HPARAMS, BEST_NEWTON_GYM = scan_hparams_newton()

CROSS_VAL_GYM = cross_validate_newton(BEST_HPARAMS, BEST_NEWTON_GYM)

plot_everything(BEST_STEP_SIZE, BEST_GYM, BEST_HPARAMS, BEST_NEWTON_GYM)

# ### Conclusions
#
# - The NEWTON and ADAGRAD optimizers are more efficient to adapt to slowly changing environments.
# - The SGD and ADAM optimizers seem to have the worst performance.

# ### Fixed setting

# %%time
run_fixed_setting()


# ## Setting 3

# ### Environment

# Let us build an environment corresponding to the "setting 3" of [1].
# We modify it slightly by adding 10000 steps. We intentionally use
# use the 5000 to 10000 steps to optimize the hyper parameters.
# This allows us to evaluate how the models "over-optimize".

# +
from wax.modules import Counter


def build_env():
    def env(action, obs):
        y_pred, eps = action, obs
        t = Counter()()
        ar_coefs_1 = jnp.array([0.6, -0.5, 0.4, -0.4, 0.3])
        ar_coefs_2 = jnp.array([-0.4, -0.5, 0.4, 0.4, 0.1])

        ar_coefs = jnp.where(t < int(Tlong / 2), ar_coefs_1, ar_coefs_2)
        ma_coefs_1 = jnp.array([0.3, -0.2])
        ma_coefs_2 = jnp.array([-0.3, 0.2])
        ma_coefs = jnp.where(t < int(Tlong / 2), ma_coefs_1, ma_coefs_2)

        y = ARMA(ar_coefs, ma_coefs)(eps)

        # prediction used on a fresh y observation.
        rw = -((y - y_pred) ** 2)

        env_info = {"y": y, "y_pred": y_pred}
        obs = y
        return rw, obs, env_info

    return env


def sample_noise(rng):
    eps = jax.random.uniform(rng, (Tlong, N_BATCH), minval=-0.5, maxval=0.5)
    return eps


Tlong = 2 * T
MIN_ERR = 0.0833
MAX_ERR = 0.12
LEARN_TIME_SLICE = slice(int(T / 2), T)
env = build_env()
# -

BEST_STEP_SIZE, BEST_GYM = scan_hparams_first_order()

CROSS_VAL_GYM = cross_validate_first_order(BEST_STEP_SIZE, BEST_GYM)

BEST_HPARAMS, BEST_NEWTON_GYM = scan_hparams_newton()

CROSS_VAL_GYM = cross_validate_newton(BEST_HPARAMS, BEST_NEWTON_GYM)

plot_everything(BEST_STEP_SIZE, BEST_GYM, BEST_HPARAMS, BEST_NEWTON_GYM)

# ### Choosing hyper parameters on the whole period

# It seems that Newton solver is more prone to overfitting (recall that we chose its hyper parameters to optimize the average loss between steps 5000 and 1000, thus only in the first regime).
#
#
# However, as stated in [1], Newton algorithm can have better performances if we choose its hyper parameters in order to obtain the best performances for both regimes.
#
# Let us check this:

LEARN_TIME_SLICE = slice(int(Tlong / 2), None)
env = build_env()

BEST_STEP_SIZE, BEST_GYM = scan_hparams_first_order()

CROSS_VAL_GYM = cross_validate_first_order(BEST_STEP_SIZE, BEST_GYM)

BEST_HPARAMS, BEST_NEWTON_GYM = scan_hparams_newton()

CROSS_VAL_GYM = cross_validate_newton(BEST_HPARAMS, BEST_NEWTON_GYM)

plot_everything(BEST_STEP_SIZE, BEST_GYM, BEST_HPARAMS, BEST_NEWTON_GYM)

# ### Conclusion
#
# - The ADAGRAD optimizers seems to be best suited for abrupt regime switching.
# - The SGD and NEWTON optimizers seem to behave similarly if their parameters are correctly chosen.
# - The ADAM optimizer seems to have the worst performance.

# ### Fixed setting

# %%time
run_fixed_setting()


# ## Setting 4

# ### Environment

# let's build an environment corresponding to "setting 4" in [1]

# +
from wax.modules import Counter


def build_env():
    def env(action, obs):
        y_pred, eps = action, obs
        t = Counter()()
        ar_coefs = jnp.array([0.11, -0.5])

        ma_coefs = jnp.array([0.41, -0.39, -0.685, 0.1])

        # rng = hk.next_rng_key()

        prev_eps = hk.get_state("prev_eps", (1,), init=lambda *_: jnp.zeros_like(eps))
        eps = prev_eps + eps  # jax.random.normal(rng, (1, N_BATCH))

        hk.set_state("prev_eps", eps)

        y = ARMA(ar_coefs, ma_coefs)(eps)
        # prediction used on a fresh y observation.
        rw = -((y - y_pred) ** 2)

        env_info = {"y": y, "y_pred": y_pred}
        obs = y
        return rw, obs, env_info

    return env


def sample_noise(rng):
    eps = jax.random.normal(rng, (T, N_BATCH)) * 0.3
    return eps


MIN_ERR = 0.09
MAX_ERR = 0.3
LEARN_TIME_SLICE = slice(int(T / 2), T)
env = build_env()
# -


BEST_STEP_SIZE, BEST_GYM = scan_hparams_first_order()

BEST_STEP_SIZE

CROSS_VAL_GYM = cross_validate_first_order(BEST_STEP_SIZE, BEST_GYM)

BEST_HPARAMS, BEST_NEWTON_GYM = scan_hparams_newton()

CROSS_VAL_GYM = cross_validate_newton(BEST_HPARAMS, BEST_NEWTON_GYM)

plot_everything(BEST_STEP_SIZE, BEST_GYM, BEST_HPARAMS, BEST_NEWTON_GYM)

# As noted in [1], the newton algorithm seems to be the only one to achieve an average error rate that converges to the variance of the noise (0.09).
#

# ### Conclusion
#
# In this environment with noise auto-correlations:
# - The NEWTON optimizer achieve to realize the minimum theoretical average loss
# - The other optimizers struggle to converge to the minimum theoretical loss and thus seems to suffer a linear regret.
# - The SGD optimizer is the worst in this setting.

# ### Fixed setting

# %%time
run_fixed_setting()



# Copyright 2022 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Optional

import haiku as hk
import jax
import jax.numpy as jnp

from wax.modules import FillNanInf
from wax.modules.update_params import UpdateParams, get_init_params
from wax.predicate import pass_all_predicate


class FuncOptimizer(hk.Module):
    """Optimize a function with an iterative solver.

    Args:
        func: a function to optimize
        opt: an iterative solver
        has_aux: option propagated to jax.grad
        params_predicate: predicate function to partition parameters in trainable parameter
            and non-trainable parameters.
        grads_fill_nan_inf: if True, fill nan and infinite value of the gradients with zeros.
        name: name of the module.

    Returns:
        results: return of func
        trainable_params: updated trainable params
    """

    def __init__(
        self,
        func,
        opt,
        has_aux=False,
        params_predicate: Optional[
            Callable[[str, str, jnp.ndarray], bool]
        ] = pass_all_predicate,
        grads_fill_nan_inf=False,
        name=None,
    ):
        self.func = func
        self.opt = opt
        self.has_aux = has_aux
        self.grads_fill_nan_inf = grads_fill_nan_inf
        self.params_predicate = params_predicate
        super().__init__(name=name)

    def __call__(self, *args, **kwargs):
        trainable_params = hk.get_state(
            "trainable_params",
            [],
            init=lambda *_: get_init_params(
                self.func, params_predicate=self.params_predicate, *args, **kwargs
            ),
        )
        func = UpdateParams(self.func, params_predicate=self.params_predicate)
        results, grads = jax.value_and_grad(func, has_aux=self.has_aux)(
            trainable_params, *args, **kwargs
        )
        if self.grads_fill_nan_inf:
            grads = FillNanInf()(grads)

        # trainable_params = jax.tree_util.tree_map(self.opt, trainable_params, grads)
        trainable_params = self.opt(trainable_params, grads)

        hk.set_state("trainable_params", trainable_params)
        return results, trainable_params



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp

from wax.modules import EWMA, VMap
from wax.stateful import unroll_lift_with_state
from wax.unroll import unroll


def test_vmap_with_outer_unroll_lift_with_state():
    """Test that VMap module work without PRNG key specified"""
    x = jnp.arange(2 * 2 * 2).reshape(2, 2, 2).astype(jnp.float32)

    def outer_fun(x):
        @unroll_lift_with_state
        def step(x):
            return VMap(EWMA(com=10))(x)

        return step(x)

    res = unroll(outer_fun, rng=jax.random.PRNGKey(0))(x)
    assert res.shape == x.shape
    ref = jnp.array(
        [
            [[0.0, 1.0], [1.0476191, 2.047619]],
            [[2.1268883, 3.1268883], [3.2376645, 4.2376647]],
        ],
        dtype=jnp.float32,
    )
    assert jnp.allclose(res, ref)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute exponentioal moving average."""
from dataclasses import dataclass
from typing import Any, NamedTuple, Optional, cast

import numba
import numpy as np
import pandas as pd


class EWMAState(NamedTuple):
    mean: Any
    old_wt: Any
    nobs: Any


def ewma(
    *,
    com: Optional[float] = None,
    alpha: Optional[float] = None,
    min_periods: int = 0,
    adjust: bool = True,
    ignore_na: bool = False,
    initial_value=np.nan,
):
    r"""Compute exponentioal moving average.

    Args:
        alpha:  Specify smoothing factor :math:`\alpha` directly
            :math:`0 < \alpha \leq 1`.
        com : Specify decay in terms of center of mass
            :math:`\alpha = 1 / (1 + com)`, for :math:`com \geq 0`.

        min_periods : Minimum number of observations in window required to have a value;
            otherwise, result is ``np.nan``.


        adjust : Divide by decaying adjustment factor in beginning periods to account
            for imbalance in relative weightings (viewing EWMA as a moving average).
            - When ``adjust=True`` (default), the EW function is calculated using weights
              :math:`w_i = (1 - \alpha)^i`. For example, the EW moving average of the series
              [:math:`x_0, x_1, ..., x_t`] would be:
            .. math::
                y_t = \frac{x_t + (1 - \alpha)x_{t-1} + (1 - \alpha)^2 x_{t-2} + ... + (1 -
                \alpha)^t x_0}{1 + (1 - \alpha) + (1 - \alpha)^2 + ... + (1 - \alpha)^t}
            - When ``adjust=False``, the exponentially weighted function is calculated
              recursively:
            .. math::
                \begin{split}
                    y_0 &= x_0\\
                    y_t &= (1 - \alpha) y_{t-1} + \alpha x_t,
                \end{split}
            The effective  center of mass (com) interpolate exponentially between 0 and the
            nominal center of mass.

            - When ``adjust='linear'`` the effective  center of mass (com) interpolate linearly
            between 0 and the nominal center of mass.

        ignore_na : Ignore missing values when calculating weights.
            - When ``ignore_na=False`` (default), weights are based on absolute positions.
              For example, the weights of :math:`x_0` and :math:`x_2` used in calculating
              the final weighted average of [:math:`x_0`, None, :math:`x_2`] are
              :math:`(1-\alpha)^2` and :math:`1` if ``adjust=True``, and
              :math:`(1-\alpha)^2` and :math:`\alpha` if ``adjust=False``.
            - When ``ignore_na=True``, weights are based
              on relative positions. For example, the weights of :math:`x_0` and :math:`x_2`
              used in calculating the final weighted average of
              [:math:`x_0`, None, :math:`x_2`] are :math:`1-\alpha` and :math:`1` if
              ``adjust=True``, and :math:`1-\alpha` and :math:`\alpha` if ``adjust=False``.


        initial_value : initial value for the state.
    """
    assert (
        com is not None or alpha is not None
    ), "com or alpha parameters must be specified."
    if com is not None:
        assert alpha is None
    elif alpha is not None:
        assert com is None
        com = 1.0 / alpha - 1.0
    com = cast(float, com)
    assert com > 0.0
    alpha = 1.0 / (1.0 + com)

    def init_ewma_state(x):
        x = x[0]

        dtype = x.dtype
        shape = x.shape

        state = EWMAState(
            mean=np.full(shape, initial_value, dtype),
            old_wt=np.full(shape, 1.0, dtype),
            nobs=np.full(shape, 0.0, dtype=np.int64),
        )
        return state

    def apply(values: np.ndarray, state: Optional[EWMAState] = None):
        is_1d = False
        if values.ndim == 1:
            values = values.reshape(-1, 1)
            is_1d = True

        if state is None:
            state = init_ewma_state(values)
        mean = state.mean
        old_wt = state.old_wt
        nobs = state.nobs

        res, mean, old_wt, nobs = numba_apply(values, mean, old_wt, nobs)
        state = EWMAState(mean, old_wt, nobs)

        if is_1d:
            res = res.reshape(values.shape[0])
        return res, state

    @numba.jit(nopython=True, nogil=True, parallel=False)
    def numba_apply(values, mean, old_wt, nobs):

        """
        Compute online exponentially weighted mean per column over 2D values.

        Takes the first observation as is, then computes the subsequent
        exponentially weighted mean accounting minimum periods.
        """
        minimum_periods = min_periods

        if adjust:
            new_wt = 1.0
        else:
            new_wt = alpha

        # deltas = np.ones(values.shape)
        result = np.empty(values.shape)
        weighted_avg = mean

        for i in range(len(values)):
            cur = values[i]
            is_observations = ~np.isnan(cur)
            nobs += is_observations.astype(np.int64)
            for j in numba.prange(len(cur)):
                if not np.isnan(weighted_avg[j]):
                    if adjust == "linear":
                        if is_observations[j]:
                            old_wt_factor = 1.0
                        else:
                            if old_wt[j] > 0:
                                old_wt_factor = np.maximum(
                                    0.0, (old_wt[j] - 1.0) / old_wt[j]
                                )
                            else:
                                old_wt_factor = 0.0
                        old_wt[j] = np.minimum(old_wt[j], com)
                    else:
                        old_wt_factor = 1.0 - alpha

                    if is_observations[j] or not ignore_na:
                        # note that len(deltas) = len(vals) - 1 and deltas[i] is to be
                        # used in conjunction with vals[i+1]
                        old_wt[j] *= old_wt_factor  # ** deltas[j - 1]
                        if is_observations[j]:
                            # avoid numerical errors on constant series
                            if weighted_avg[j] != cur[j]:
                                weighted_avg[j] = (
                                    (old_wt[j] * weighted_avg[j]) + (new_wt * cur[j])
                                ) / (old_wt[j] + new_wt)
                            if adjust:
                                old_wt[j] += new_wt
                            else:
                                old_wt[j] = 1.0
                elif is_observations[j]:
                    weighted_avg[j] = cur[j]

            result[i] = np.where(nobs >= minimum_periods, weighted_avg, np.nan)
            mean = weighted_avg
        return result, mean, old_wt, nobs

    return apply


class WaxAccessor:
    def ewm(self, *args, **kwargs):
        return NumbaExponentialMovingWindow(accessor=self, *args, **kwargs)


@dataclass(frozen=True)
class NumbaExponentialMovingWindow:
    accessor: WaxAccessor
    com: Optional[float] = None
    alpha: Optional[float] = None
    min_periods: int = 0
    adjust: bool = True
    ignore_na: bool = False
    initial_value: float = np.nan
    return_state: bool = False

    def mean(self, state=None):
        def _apply_ema(
            *,
            state,
            obj,
            com,
            alpha,
            min_periods,
            adjust,
            ignore_na,
            initial_value,
            return_state,
        ):
            if state is not None:
                state = state.reindex(obj.columns)
                state = EWMAState(
                    mean=state["mean"].values,
                    old_wt=state["old_wt"].values,
                    nobs=state["nobs"].values,
                )

            res, state = ewma(
                alpha=alpha,
                com=com,
                min_periods=min_periods,
                adjust=adjust,
                ignore_na=ignore_na,
                initial_value=initial_value,
            )(obj.values, state)

            res = pd.DataFrame(res, obj.index, obj.columns)

            state = state._replace(
                mean=pd.Series(state.mean, obj.columns),
                old_wt=pd.Series(state.old_wt, obj.columns),
                nobs=pd.Series(state.nobs, obj.columns),
            )
            state = pd.concat(state._asdict(), axis=1)
            if return_state:
                return res, state
            else:
                return res

        kwargs = self.__dict__.copy()
        obj = kwargs.pop("accessor")._obj

        if isinstance(obj, pd.Series):
            kwargs["obj"] = obj.to_frame()
            res = _apply_ema(state=state, **kwargs)
            if self.return_state:
                res, state = res
            res = res.iloc[:, 0]
            if self.return_state:
                return res, state
            else:
                return res
        else:
            kwargs["obj"] = obj
            return _apply_ema(state=state, **kwargs)


class WaxNumbaEWMAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def ewm(self, *args, **kwargs):
        return NumbaExponentialMovingWindow(accessor=self, *args, **kwargs)


def register_wax_numba():
    pd.api.extensions.register_dataframe_accessor("wax_numba")(WaxNumbaEWMAccessor)
    pd.api.extensions.register_series_accessor("wax_numba")(WaxNumbaEWMAccessor)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
b"""## Callbacks

We now develop a callbacks API in order to simplify as much as possible the local
training/simulation loop.

It permits to implement all complex stuffs in separated code that we call callbacks.
This can be to:
- record results
- compute additional metrics
- adding some monitoring tools: ProgressBar, ...
- control workflow: Stop iteration, ...

The design of the callback API bellow is highly inspired from the one used internally in the
library `dask` (see for ex `ProgressBar` callback).
"""

from contextlib import contextmanager
from typing import Set


class Callback:
    __slots__ = (
        "_cm",
        "_on_train_start",
        "_on_act",
        "_on_step",
        "_on_train_start",
        "_on_train_end",
    )
    """Base class for using the callback mechanism

    Create a callback with functions of the following signatures:

    >>> def _on_train_start(env, agent, obs):
    ...     pass
    >>> def on_act(env, agent, obs, action):
    ...     pass
    >>> def on_step(env, agent, gym_state):
    ...     pass
    >>> def _on_train_end(env, agent):
    ...     pass

    You may then construct a callback object with any number of them

    >>> cb = Callback(on_step=on_step) # doctest: +SKIP

    And use it either as a context manager over a compute/get call

    >>> with cb:            # doctest: +SKIP
    ...     gym_train(agent, env)     # doctest: +SKIP

    Or globally with the ``register`` method

    >>> cb.register()       # doctest: +SKIP
    >>> cb.unregister()     # doctest: +SKIP

    Alternatively subclass the ``Callback`` class with your own methods.

    >>> class PrintReward(Callback):      # doctest: +SKIP
    ...     def __init__(self): pass
    ...     def _on_step(env, agent, gym_state):
    ...         print(f"Reward = {gym_state.rw:.4g}")

    >>> with PrintReward():   # doctest: +SKIP
    ...     gym_train(agent, env)     # doctest: +SKIP
    """

    active: Set = set()

    def __init__(
        self, on_train_start=None, on_act=None, on_step=None, on_train_end=None
    ):
        if on_train_start:
            self._on_train_start = on_train_start
        if on_act:
            self._on_act = on_act
        if on_step:
            self._on_step = on_step
        if on_train_end:
            self._on_train_end = on_train_end

    @property
    def _callback(self):
        fields = ["_on_train_start", "_on_act", "_on_step", "_on_train_end"]
        return tuple(getattr(self, i, None) for i in fields)

    def __enter__(self):
        self._cm = add_callbacks(self)
        self._cm.__enter__()
        return self

    def __exit__(self, *args):
        self._cm.__exit__(*args)

    def register(self):
        Callback.active.add(self._callback)

    def unregister(self):
        Callback.active.remove(self._callback)


def unpack_callbacks(cbs):
    """Take an iterable of callbacks, return a list of each callback."""
    if cbs:
        return [[i for i in f if i] for f in zip(*cbs)]
    else:
        return [(), (), (), ()]


@contextmanager
def local_callbacks(callbacks=None):
    """Allows callbacks to work with nested schedulers.

    Callbacks will only be used by the first started scheduler they encounter.
    This means that only the outermost scheduler will use global callbacks."""
    global_callbacks = callbacks is None
    if global_callbacks:
        callbacks, Callback.active = Callback.active, set()
    try:
        yield callbacks or ()
    finally:
        if global_callbacks:
            Callback.active = callbacks


def normalize_callback(cb):
    """Normalizes a callback to a tuple"""
    if isinstance(cb, Callback):
        return cb._callback
    elif isinstance(cb, tuple):
        return cb
    else:
        raise TypeError("Callbacks must be either `Callback` or `tuple`")


class add_callbacks(object):
    """Context manager for callbacks.

    Takes several callbacks and applies them only in the enclosed context.
    Callbacks can either be represented as a ``Callback`` object, or as a tuple
    of length 4.

    Examples
    --------
    >>> def pretask(key, dsk, state):
    ...     print("Now running {0}").format(key)
    >>> callbacks = (None, pretask, None, None)
    >>> with add_callbacks(callbacks):    # doctest: +SKIP
    ...     res.compute()
    """

    def __init__(self, *callbacks):
        self.callbacks = [normalize_callback(c) for c in callbacks]
        Callback.active.update(self.callbacks)

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        for c in self.callbacks:
            Callback.active.discard(c)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp
import optax

from wax.optim.newton import newton, sherman_morrison


def test_sherman_morrisson():
    x = 1
    A_inv = jnp.eye(3) * 1.0
    u = v = jnp.ones(3)

    ref = jnp.linalg.inv(A_inv + x * jnp.outer(u, v))

    A_inv = sherman_morrison(A_inv, u, v)
    assert jnp.allclose(A_inv, ref)


def loss(w, x):
    return -(w * x).sum() + (w**2).sum()


# from optax import sgd, adagrad


def test_sgd():

    w = jnp.ones(3)
    x = jnp.ones(3)

    l_ = loss(w, x)

    # opt = sgd(1.0e-3)
    # opt = adagrad(1.0e-3)
    opt = newton(1.0e-3)

    l_, grads = jax.value_and_grad(loss)(w, x)

    state = opt.init(w)
    grads, state = opt.update(grads, state)
    w = optax.apply_updates(w, grads)

    assert loss(w, x) < l_



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as onp
import pandas as pd
import pytest
from jax.config import config

from wax.compile import jit_init_apply
from wax.modules.ewma import EWMA
from wax.unroll import dynamic_unroll_fori_loop, unroll, unroll_transform_with_state


@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_init_and_first_step_ema(dtype):

    if dtype == "float64":
        config.update("jax_enable_x64", True)
    else:
        config.update("jax_enable_x64", False)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(3,), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(alpha=0.1, adjust=True)(x)

    params, state = model.init(next(seq), x)
    ema, state = model.apply(params, state, next(seq), x)
    assert ema.dtype == jnp.dtype(dtype)


def test_run_ema_vs_pandas_not_adjust():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(alpha=0.1, adjust=False)(x)

    ema, state = unroll(model, dynamic=False, return_final_state=True)(x)

    pandas_ema = pd.DataFrame(x).ewm(alpha=0.1, adjust=False).mean()

    assert jnp.allclose(ema, pandas_ema.values)


def test_dynamic_unroll_fori_loop():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(alpha=0.1, adjust=True)(x)

    ema, state = unroll(model, dynamic=False, return_final_state=True)(x)

    ema2, state2 = dynamic_unroll_fori_loop(model, None, None, next(seq), False, x)

    assert jnp.allclose(ema, ema2)


def test_dynamic_unroll():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(alpha=0.1, adjust=True)(x)

    ema, state = unroll(model, dynamic=False, return_final_state=True)(x)

    seq = hk.PRNGSequence(42)
    ema2, state2 = unroll(model, return_final_state=True)(x)

    assert jnp.allclose(ema, ema2)


def test_run_ema_vs_pandas_adjust():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(alpha=0.1, adjust=True)(x)

    ema, state = unroll(model, return_final_state=True)(x)

    pandas_ema = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).mean()
    assert jnp.allclose(ema, pandas_ema.values)


def test_run_ema_vs_pandas_adjust_finite():
    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(alpha=0.1, adjust="linear")(x)

    ema, state = unroll(model, return_final_state=True)(x)
    pandas_ema_adjust = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).mean()
    pandas_ema_not_adjust = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).mean()
    assert not jnp.allclose(ema, pandas_ema_adjust.values)
    assert not jnp.allclose(ema, pandas_ema_not_adjust.values)
    corr = jnp.corrcoef(ema.flatten(), pandas_ema_adjust.values.flatten())[0, 1]
    assert 1.0e-3 < 1 - corr < 1.0e-2


@pytest.mark.parametrize("adjust", [False, True, "linear"])
def test_grad_ewma(adjust):

    rng = jax.random.PRNGKey(42)
    x = jax.random.normal(rng, (10, 3))
    # put some nan values
    x = x.at[0].set(jnp.nan)

    _, rng = jax.random.split(rng)

    @partial(unroll_transform_with_state, dynamic=False)
    def fun(x):
        return EWMA(alpha=1 / 10, adjust=adjust)(x)

    params, state = fun.init(rng, x)
    res, final_state = fun.apply(params, state, rng, x)

    @jax.value_and_grad
    def batch(params):
        res, final_state = fun.apply(params, state, rng, x)
        return res.mean()

    score, grad = batch(params)
    assert not jnp.isnan(grad["ewma"]["logcom"])


def check_against_pandas_ewm(x, **ewma_kwargs):
    @partial(unroll_transform_with_state, dynamic=True)
    def fun(x):
        return EWMA(return_info=True, **ewma_kwargs)(x)

    rng = jax.random.PRNGKey(42)
    params, state = fun.init(rng, x)
    (res, info), final_state = fun.apply(params, state, rng, x)
    res = pd.DataFrame(onp.array(res))

    ref_res = pd.DataFrame(onp.array(x)).ewm(**ewma_kwargs).mean()
    pd.testing.assert_frame_equal(res, ref_res, atol=1.0e-6)

    @jax.value_and_grad
    def batch(params):
        (res, info), final_state = fun.apply(params, state, rng, x)
        return jnp.nanmean(res)

    score, grad = batch(params)
    assert not jnp.isnan(grad["ewma"]["logcom"])


@pytest.mark.parametrize(
    "adjust, ignore_na",
    [(False, False), (False, True), (True, False), (True, True)],  # ,
)
def test_nan_at_beginning(adjust, ignore_na):
    config.update("jax_enable_x64", True)

    T = 20
    x = jnp.full((T,), jnp.nan).at[2].set(1).at[10].set(-1)
    check_against_pandas_ewm(x, com=10, adjust=adjust, ignore_na=ignore_na)

    # check min_periods option with random variable
    rng = jax.random.PRNGKey(42)
    x = jax.random.normal(rng, (5,))
    check_against_pandas_ewm(
        x,
        com=10,
        adjust=adjust,
        ignore_na=ignore_na,
        min_periods=2,
    )

    # check random variable with nans
    # rng = jax.random.PRNGKey(42)
    # x = jax.random.normal(rng, (6,)).at[3].set(jnp.nan)
    # x = jnp.ones((6,), "float64").at[0].set(-1).at[3].set(jnp.nan)
    x = jnp.ones((30,), "float64").at[0].set(-1).at[5:20].set(jnp.nan)
    check_against_pandas_ewm(
        x,
        com=10,
        adjust=adjust,
        ignore_na=ignore_na,
    )


def test_init_value():
    x = (
        jnp.ones((30,), "float64")
        .at[0]
        .set(jnp.nan)
        .at[1]
        .set(-1)
        .at[5:20]
        .set(jnp.nan)
    )

    res = unroll(lambda x: EWMA(com=10, adjust=False, ignore_na=False)(x))(x)
    res_init0 = unroll(
        lambda x: EWMA(com=10, adjust=False, ignore_na=False, initial_value=0.0)(x)
    )(x)

    assert res_init0[0] == 0
    assert jnp.isnan(res[0])
    assert jnp.linalg.norm(res_init0) < jnp.linalg.norm(jnp.nan_to_num(res))


def test_train_ewma():
    import optax
    from tqdm.auto import tqdm

    COM_INIT = 100
    COM_TARGET = 10
    T = 1000
    n_epochs = 100

    def train():
        def model(x):
            return EWMA(com=COM_INIT, adjust=True, ignore_na=False)(x)

        @jax.jit
        def loss(y, y_ref):
            return jnp.nanmean((y - y_ref) ** 2)

        @jax.jit
        def loss_p(params, state, x, y_ref):
            y_pred, state = model.apply(params, state, rng, x)
            return loss(y_pred, y_ref)

        rng = jax.random.PRNGKey(42)
        x = jax.random.normal(rng, (T,)).at[T // 2].set(jnp.nan)

        y_ref = unroll(lambda x: EWMA(com=COM_TARGET, adjust=True, ignore_na=False)(x))(
            x
        )

        model = unroll_transform_with_state(model)
        params, state = model.init(rng, x)

        y_pred, _ = model.apply(params, state, rng, x)

        opt = optax.adagrad(1.0e-1)
        opt_state = opt.init(params)
        losses = []
        for e in tqdm(range(n_epochs)):
            # params, state = model.init(rng, x)
            _, state = model.init(rng, x)
            y_pred, state = model.apply(params, state, rng, x)
            l_ = loss(y_pred, y_ref)
            grad = jax.grad(loss_p)(params, state, x, y_ref)
            logcom = params["ewma"]["logcom"]
            if e % 100 == 0:
                print(
                    f"e={e}, logcom={logcom}, com={jnp.exp(logcom)}, grad={grad['ewma']['logcom']}, loss = {l_}"
                )
            updates, opt_state = opt.update(grad, opt_state)
            params = optax.apply_updates(params, updates)
            losses.append(l_)
        return jnp.array(losses)

    losses = train()
    assert losses[-1] < losses[0]



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from typing import Callable, Union

import haiku as hk
import jax
from haiku import TransformedWithState

from wax.stateful import vmap_lift_with_state


class VMap(hk.Module):
    def __init__(self, fun: Callable, name=None):
        """A Haiku module that applies a function to a batch of inputs."""
        warnings.warn(
            "VMap is deprecated, please use vmap_lift_with_state instead.",
            DeprecationWarning,
        )
        super().__init__(name=name)
        self.fun = fun

    def __call__(self, *args, **kwargs):
        return vmap_lift_with_state(self.fun, split_rng=False, init_rng=True)(
            *args, **kwargs
        )


# Helper functions


def add_batch(fun: Union[Callable, TransformedWithState], take_mean=True):
    """Wrap a function with vmap_lift_with_state.
    It should be used inside a transformed function."""

    def fun_batch(*args, **kwargs):
        try:
            res = vmap_lift_with_state(fun, split_rng=True)(*args, **kwargs)
        except ValueError:
            res = vmap_lift_with_state(fun, split_rng=False)(*args, **kwargs)

        if take_mean:
            res = jax.tree_util.tree_map(lambda x: x.mean(axis=0), res)
        return res

    return fun_batch



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unroll agent with a gym environment."""
from collections import namedtuple

from wax.gym.callbacks.callbacks import local_callbacks, unpack_callbacks


def gym_unroll(agent, env, obs=None, callbacks=None):
    """Unroll agent with a gym environment.

    Parameters
    ----------
    env
        Gym environment used to unroll the data and feed the agent.

    obs
        if passed, start the Gym unroll loop with this observation instead of
        the observation returned by 'env.reset()', else 'obs = env.reset()' is
        used as first observation.

    callbacks
        Additional callback to use

    Returns
    -------

    gym_state
        nested data structure
    """
    GymUnrollState = namedtuple("GymUnrollState", "rw, obs, action, done, info")

    with local_callbacks(callbacks) as callbacks:

        # unpack callbacks
        (
            on_train_start_cbs,
            on_act_cbs,
            on_step_cbs,
            on_train_end_cbs,
        ) = unpack_callbacks(callbacks)

        obs = env.reset() if obs is None else obs

        # call "on_train_start" callbacks
        for on_train_start in on_train_start_cbs:
            on_train_start(env, agent, obs)

        while True:

            # agent act
            action = agent(obs)

            # call "on_act" callbacks
            for on_act in on_act_cbs:
                on_act(env, agent, obs, action)

            # env step
            obs, rw, done, info = env.step(action)

            # set loop state
            gym_state = GymUnrollState(rw, obs, action, done, info)

            # call "on_step" callbacks
            for on_step in on_step_cbs:
                step_done = on_step(env, agent, gym_state)
                done = done or step_done

            if done:
                break

        # call "on_train_end" callbacks
        for on_train_end in on_train_end_cbs:
            on_train_end(env, agent)

        env.close()

    return gym_state



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute mean over masked data."""

import haiku as hk
import jax.numpy as jnp
from jax import tree_map

from wax.modules.apply_mask import ApplyMask


class MaskMean(hk.Module):
    """Compute mean over masked data."""

    def __init__(self, axis=None, name=None):
        """Initialize the module.

        Args:
            axis : axis along which to compute the mean
        """
        super().__init__(name=name)
        self.axis = axis

    def __call__(self, mask, input):
        """Compute mean.

        Args:
            mask : bolean mask specifying points over which to compute the mean.
            input : data on which to compute the mean.
        """

        def mask_mean(x):
            # mask values
            x = ApplyMask()(mask, x)

            # fill nans
            x = jnp.nan_to_num(x)

            count = mask.sum(axis=self.axis)
            mean = x.sum(axis=self.axis) / count
            return mean

        return tree_map(mask_mean, input)



# Copyright 2022 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import optax


class OptaxOptimizer(hk.Module):
    """A module which wraps an optax GrandientTransformation.

    Args:
        opt: gradient transformation
        name: name of the module
    """

    def __init__(self, opt: optax.GradientTransformation, name=None):
        super().__init__(name=name)
        self.opt = opt

    def __call__(self, params, grads):
        opt_state = hk.get_state("opt_state", [], init=lambda *_: self.opt.init(params))
        updates, opt_state = self.opt.update(grads, opt_state)
        params = optax.apply_updates(params, updates)
        hk.set_state("opt_state", opt_state)
        return params



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from wax.modules.apply_mask import ApplyMask
from wax.modules.arma import ARMA
from wax.modules.buffer import Buffer
from wax.modules.counter import Counter
from wax.modules.diff import Diff
from wax.modules.ewma import EWMA
from wax.modules.ewmcov import EWMCov
from wax.modules.ewmvar import EWMVar
from wax.modules.ffill import Ffill
from wax.modules.fill_nan_inf import FillNanInf
from wax.modules.gym_feedback import GymFeedback
from wax.modules.has_changed import HasChanged
from wax.modules.lag import Lag
from wax.modules.mask_mean import MaskMean
from wax.modules.mask_normalize import MaskNormalize
from wax.modules.mask_std import MaskStd
from wax.modules.ohlc import OHLC
from wax.modules.online_optimizer import OnlineOptimizer
from wax.modules.online_supervised_learner import OnlineSupervisedLearner
from wax.modules.pct_change import PctChange
from wax.modules.rolling_mean import RollingMean
from wax.modules.snarimax import SNARIMAX
from wax.modules.update_on_event import UpdateOnEvent
from wax.modules.update_params import UpdateParams
from wax.modules.vmap import VMap

__all__ = [
    "Buffer",
    "Diff",
    "EWMA",
    "EWMCov",
    "EWMVar",
    "GymFeedback",
    "HasChanged",
    "Lag",
    "OHLC",
    "Ffill",
    "PctChange",
    "RollingMean",
    "UpdateOnEvent",
    "FillNanInf",
    "ApplyMask",
    "MaskMean",
    "MaskStd",
    "MaskNormalize",
    "OnlineSupervisedLearner",
    "VMap",
    "UpdateParams",
    "ARMA",
    "SNARIMAX",
    "OnlineOptimizer",
    "Counter",
]



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.3
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# +
# Uncomment to run the notebook in Colab
# # ! pip install -q "wax-ml[complete]@git+https://github.com/eserie/wax-ml.git"
# # ! pip install -q --upgrade jax jaxlib==0.1.70+cuda111 -f https://storage.googleapis.com/jax-releases/jax_releases.html
# -

# check available devices
import jax

print("jax backend {}".format(jax.lib.xla_bridge.get_backend().platform))
jax.devices()

# # 🦎 Online linear regression with a non-stationary environment 🦎
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/eserie/wax-ml/blob/main/docs/notebooks/06_Online_Linear_Regression.ipynb)

# We implement an online learning non-stationary linear regression problem.
#
# We go there progressively by showing how a linear regression problem can be cast
# into an online learning problem thanks to the `OnlineSupervisedLearner` module.
#
# Then, to tackle a non-stationary linear regression problem (i.e. with a weight that can vary in time)
# we reformulate the problem into a reinforcement learning problem that we implement with the `GymFeedBack` module of WAX-ML.
#
# We then need to define an "agent" and an "environment" using simple functions implemented with modules:
# - The agent is responsible for learning the weights of its internal linear model.
# - The environment is responsible for generating labels and evaluating the agent's reward metric.
#
# We experiment with a non-stationary environment that returns the sign of the linear regression parameters at a given time step,
# known only to the environment.
#
# This example shows that it is quite simple to implement this online-learning task with WAX-ML tools.
# In particular, the functional workflow adopted here allows reusing the functions implemented for a
# task for each new task of increasing complexity,
#
#
# In this journey, we will use:
# - Haiku basic linear module `hk.Linear`.
# - Optax stochastic gradient descent optimizer: `sgd`.
# - WAX-ML modules: `OnlineSupervisedLearner`, `Lag`, `GymFeedBack`
# - WAX-ML helper functions: `unroll`, `jit_init_apply`

# %pylab inline

import haiku as hk
import jax
import jax.numpy as jnp
import optax
from matplotlib import pyplot as plt

from wax.compile import jit_init_apply
from wax.modules import OnlineSupervisedLearner

# ## Static Linear Regression

# First, let's implement a simple linear regression

# ### Generate data

# Let's generate a batch of data:

seq = hk.PRNGSequence(42)
X = jax.random.normal(next(seq), (100, 3))
w_true = jnp.ones(3)

# ### Define the model

# We use the basic module `hk.Linear` which is a linear layer.
# By default, it initializes the weights with random values from the truncated normal,
# with a standard deviation of $1 / \sqrt{N}$ (See https://arxiv.org/abs/1502.03167v3)
# where $N$ is the size of the inputs.


@jit_init_apply
@hk.transform_with_state
def linear_model(x):
    return hk.Linear(output_size=1, with_bias=False)(x)


# ### Run the model

# Let's run the model using WAX-ML `unroll` on the batch of data.

from wax.unroll import unroll

params, state = linear_model.init(next(seq), X[0])
linear_model.apply(params, state, None, X[0])

Y_pred = unroll(linear_model, rng=next(seq))(X)

Y_pred.shape


# ### Check cost

# Let's look at the mean squared error for this non-trained model.

noise = jax.random.normal(next(seq), (100,))
Y = X.dot(w_true) + noise
L = ((Y - Y_pred) ** 2).sum(axis=1)
mean_loss = L.mean()
assert mean_loss > 0


# Let's look at the regret (cumulative sum of the loss) for the non-trained model.

plt.plot(L.cumsum())
plt.title("Regret")

# As expected, we have a linear regret when we did not train the model!

# ## Online Linear Regression


# We will now start training the model online.
# For a review on online-learning methods see [1]
#
#
# [1] [Elad Hazan, Introduction to Online Convex Optimization](https://arxiv.org/abs/1909.05207)

# ### Define an optimizer

opt = optax.sgd(1e-3)


# ### Define a loss

# Since we are doing online learning, we need to define a local loss function:
# $$
# \ell_t(y, w, x) = \lVert y_t -  w \cdot x_t \rVert^2
# $$


def linear_model(x):
    return hk.Linear(output_size=1, with_bias=False)(x)


def loss(y_pred, y):
    return jnp.mean(jnp.square(y_pred - y))


# ### Define a learning strategy


def learner(x, y):
    return OnlineSupervisedLearner(linear_model, opt, loss)(x, y)


# ### Generate data


def generate_many_observations(T=300, sigma=1.0e-2, rng=None):
    rng = jax.random.PRNGKey(42) if rng is None else rng
    X = jax.random.normal(rng, (T, 3))
    noise = sigma * jax.random.normal(rng, (T,))
    w_true = jnp.ones(3)
    noise = sigma * jax.random.normal(rng, (T,))
    Y = X.dot(w_true) + noise
    return (X, Y)


T = 3000
X, Y = generate_many_observations(T)

# ### Unroll the learner

(output, info) = unroll(learner, rng=next(seq))(X, Y)

# ### Plot the regret

# Let's look at the loss and regret over time.

# +
fig, axs = plt.subplots(1, 2, figsize=(9, 3))
axs[0].plot(info.loss.cumsum())
axs[0].set_title("Regret")

axs[1].plot(info.params["linear"]["w"][:, 0, 0])
axs[1].set_title("Weight[0,0]")


# -

# We have sub-linear regret!

# ##  Online learning with Gym
#
# Now we will recast the online linear regression learning task as a reinforcement learning task
# implemented with the `GymFeedback` module of WAX-ML.
#
# For that, we define:
# - obserbations (`obs`) : pairs  `(x, y)` of features and labels
# - raw observations (`raw_obs`): pairs `(x, noise)`  of features and noise.

# ### Linear regression agent

# In WAX-ML, an agent is a simple function with the following API:
# <div align="center">
# <img src="../tikz/agent.png" alt="logo" width="20%"></img>
# </div>

# Let's define a simple linear regression agent with the elements we have defined so far.


def linear_regression_agent(obs):
    x, y = obs

    def model(x):
        return hk.Linear(output_size=1, with_bias=False)(x)

    opt = optax.sgd(1e-3)

    def loss(y_pred, y):
        return jnp.mean(jnp.square(y_pred - y))

    return OnlineSupervisedLearner(model, opt, loss)(x, y)


# ### Linear regression environment

# In WAX-ML, an environment is a simple function with the following API:
# <div align="center">
# <img src="../tikz/env.png" alt="logo" width="20%"></img>
# </div>

# Let's now define a linear regression environment that, for the moment,
# have static weights.
#
# It is responsible for generating the real labels and evaluating the agent's reward.
#
#
# For the evaluation of the reward, we need the `Lag` module to evaluate the action of
# the agent with the labels generated in the previous time step.

from wax.modules import Lag


def stationary_linear_regression_env(action, raw_obs):

    # Only the environment now the true value of the parameters
    w_true = -jnp.ones(3)

    # The environment has its proper loss definition
    def loss(y_pred, y):
        return jnp.mean(jnp.square(y_pred - y))

    # raw observation contains features and generative noise
    x, noise = raw_obs

    # generate targets
    y = x @ w_true + noise
    obs = (x, y)

    y_previous = Lag(1)(y)
    # evaluate the prediction made by the agent
    y_pred = action
    reward = loss(y_pred, y_previous)

    return reward, obs, {}


# ### Generate raw observation
#
# Let's define a function that generate the raw observation:


def generate_many_raw_observations(T=300, sigma=1.0e-2, rng=None):
    rng = jax.random.PRNGKey(42) if rng is None else rng
    X = jax.random.normal(rng, (T, 3))
    noise = sigma * jax.random.normal(rng, (T,))
    return (X, noise)


# ### Implement Feedback

# We are now ready to set things up with the `GymFeedback` module implemented in WAX-ML.
#
# It implements the following feedback loop:
# <div align="center">
# <img src="../tikz/gymfeedback.png" alt="logo" width="50%"></img>
# </div>

# Equivalently, it can be described with the pair of `init` and `apply` functions:
# <div align="center">
# <img src="../tikz/gymfeedback_init_apply.png" alt="logo" width="100%"></img>
# </div>

from wax.modules import GymFeedback


def gym_fun(raw_obs):
    return GymFeedback(
        linear_regression_agent, stationary_linear_regression_env, return_action=True
    )(raw_obs)


# And now we can unroll it on a sequence of raw observations!

seq = hk.PRNGSequence(42)
T = 3000
raw_observations = generate_many_raw_observations(T)
rng = next(seq)
(gym_output, gym_info) = unroll(gym_fun, rng=rng, skip_first=True)(raw_observations)

# Let's visualize the outputs.
#
#
# We now use `pd.Series` to represent the reward sequence since its first value is Nan due to the use of the lag operator.

import pandas as pd

fig, axs = plt.subplots(1, 2, figsize=(9, 3))
pd.Series(gym_output.reward).cumsum().plot(ax=axs[0], title="Regret")
axs[1].plot(info.params["linear"]["w"][:, 0, 0])
axs[1].set_title("Weight[0,0]")

# ## Non-stationary environment

# Now, let's implement a non-stationary environment.
#
# We implement it so that the sign of the weight is reversed after 2000$ steps.


class NonStationaryEnvironment(hk.Module):
    def __call__(self, action, raw_obs):

        step = hk.get_state("step", [], init=lambda *_: 0)

        # Only the environment now the true value of the parameters
        # at step 2000 we flip the sign of the true parameters !
        w_true = hk.cond(
            step < 2000,
            step,
            lambda step: -jnp.ones(3),
            step,
            lambda step: jnp.ones(3),
        )

        # The environment has its proper loss definition
        def loss(y_pred, y):
            return jnp.mean(jnp.square(y_pred - y))

        # raw observation contains features and generative noise
        x, noise = raw_obs

        # generate targets
        y = x @ w_true + noise
        obs = (x, y)

        # evaluate the prediction made by the agent
        y_previous = Lag(1)(y)
        y_pred = action
        reward = loss(y_pred, y_previous)

        step += 1
        hk.set_state("step", step)

        return reward, obs, {}


# Now let's run a gym simulation to see how the agent adapt to the
# change of environment.


def gym_fun(raw_obs):
    return GymFeedback(
        linear_regression_agent, NonStationaryEnvironment(), return_action=True
    )(raw_obs)


T = 6000
raw_observations = generate_many_raw_observations(T)
rng = jax.random.PRNGKey(42)
(gym_output, gym_info) = unroll(gym_fun, rng=rng, skip_first=True)(
    raw_observations,
)

import pandas as pd

fig, axs = plt.subplots(1, 2, figsize=(9, 3))
pd.Series(gym_output.reward).cumsum().plot(ax=axs[0], title="Regret")
axs[1].plot(gym_info.agent.params["linear"]["w"][:, 0, 0])
axs[1].set_title("Weight[0,0]")
# plt.savefig("../_static/online_linear_regression_regret.png")

# It adapts!
#
# The regret first converges, then jumps on step 2000 and finally readjusts to the new regime.
#
# We see that the weights converge to the correct values in both regimes.



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute exponentioal moving average."""
from typing import Dict, Optional, Tuple, Union, cast

import haiku as hk
from jax import numpy as jnp


class EWMA(hk.Module):
    """Compute exponential moving average."""

    def __init__(
        self,
        *,
        com: Optional[float] = None,
        alpha: Optional[float] = None,
        min_periods: int = 0,
        adjust: bool = True,
        ignore_na: bool = False,
        initial_value: float = jnp.nan,
        return_info: bool = False,
        name: Optional[str] = None,
    ):
        r"""Initialize module.

        Args:
            com : Specify decay in terms of center of mass
                :math:`\alpha = 1 / (1 + com)`, for :math:`com \geq 0`.

            alpha:  Specify smoothing factor :math:`\alpha` directly
                :math:`0 < \alpha \leq 1`.

            min_periods : Minimum number of observations in window required to have a value;
                otherwise, result is ``np.nan``.

            adjust : Divide by decaying adjustment factor in beginning periods to account
                for imbalance in relative weightings (viewing EWMA as a moving average).
                - When ``adjust=True`` (default), the EW function is calculated using weights
                  :math:`w_i = (1 - \alpha)^i`. For example, the EW moving average of the series
                  [:math:`x_0, x_1, ..., x_t`] would be:
                .. math::
                    y_t = \frac{x_t + (1 - \alpha)x_{t-1} + (1 - \alpha)^2 x_{t-2} + ... + (1 -
                    \alpha)^t x_0}{1 + (1 - \alpha) + (1 - \alpha)^2 + ... + (1 - \alpha)^t}
                - When ``adjust=False``, the exponentially weighted function is calculated
                  recursively:
                .. math::
                    \begin{split}
                        y_0 &= x_0\\
                        y_t &= (1 - \alpha) y_{t-1} + \alpha x_t,
                    \end{split}
                The effective  center of mass (com) interpolate exponentially between 0 and the
                nominal center of mass.

                - When ``adjust='linear'`` the effective  center of mass (com) interpolate linearly
                between 0 and the nominal center of mass.

            ignore_na : Ignore missing values when calculating weights.
                - When ``ignore_na=False`` (default), weights are based on absolute positions.
                  For example, the weights of :math:`x_0` and :math:`x_2` used in calculating
                  the final weighted average of [:math:`x_0`, None, :math:`x_2`] are
                  :math:`(1-\alpha)^2` and :math:`1` if ``adjust=True``, and
                  :math:`(1-\alpha)^2` and :math:`\alpha` if ``adjust=False``.
                - When ``ignore_na=True``, weights are based
                  on relative positions. For example, the weights of :math:`x_0` and :math:`x_2`
                  used in calculating the final weighted average of
                  [:math:`x_0`, None, :math:`x_2`] are :math:`1-\alpha` and :math:`1` if
                  ``adjust=True``, and :math:`1-\alpha` and :math:`\alpha` if ``adjust=False``.


            initial_value : initial value for the state.

            return_info : if true, a dictionary is returned in addition to the module output which
                contains additional variables.

            name : name of the module instance.
        """
        super().__init__(name=name)
        assert (
            com is not None or alpha is not None
        ), "com or alpha parameters must be specified."
        if com is not None:
            assert alpha is None
        elif alpha is not None:
            assert com is None
            com = 1.0 / alpha - 1.0

        self.com = cast(float, com)
        self.min_periods = min_periods
        self.adjust = adjust
        self.ignore_na = ignore_na
        self.initial_value = initial_value
        self.return_info = return_info

    def __call__(
        self, x: jnp.ndarray
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]]:
        """Compute EWMA.

        Args:
            x: input data.

        Returns:
            last_mean : value of the mean

            info: A dictionnary with additionnal variables. It is returned if `return_info` is true.
        """
        info = {}

        logcom = hk.get_parameter(
            "logcom",
            shape=[],
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.array(jnp.log(self.com), dtype),
        )
        com = jnp.exp(logcom)

        mean = hk.get_state(
            "mean",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, self.initial_value, dtype),
        )

        # get status
        is_observation = ~jnp.isnan(x)
        isnan_mean = jnp.isnan(mean)

        # fillna by zero to avoid nans in gradient computations
        x = jnp.nan_to_num(x)
        mean = jnp.nan_to_num(mean)

        alpha = 1.0 / (1.0 + com)

        if self.adjust:
            new_wt = jnp.array(1.0)
        else:
            new_wt = alpha

        old_wt = hk.get_state(
            "old_wt",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, 1.0, dtype),
        )

        if self.adjust == "linear":
            # com_eff grow linearly when there is observation but
            # decrease linearly when there is nans.
            old_wt_factor = jnp.where(
                is_observation, 1.0, jnp.maximum(0.0, (old_wt - 1.0) / old_wt)
            )
            old_wt = jnp.minimum(old_wt, com)
        else:
            old_wt_factor = 1.0 - alpha

        if self.ignore_na:
            old_wt = jnp.where(is_observation, old_wt * old_wt_factor, old_wt)
        else:
            old_wt = old_wt * old_wt_factor

        old_wt = jnp.where(isnan_mean, 0.0, old_wt)

        mean = jnp.where(
            is_observation, (old_wt * mean + new_wt * x) / (old_wt + new_wt), mean
        )

        if self.return_info:
            info["com_eff"] = old_wt / new_wt

        if self.adjust:
            old_wt = jnp.where(is_observation, old_wt + new_wt, old_wt)
        else:
            old_wt = jnp.where(is_observation, 1.0, old_wt)

        # restore nan
        mean = jnp.where(jnp.logical_and(~is_observation, isnan_mean), jnp.nan, mean)

        hk.set_state("old_wt", old_wt)
        hk.set_state("mean", mean)

        nobs = hk.get_state(
            "nobs",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, 0, dtype=int),
        )
        nobs = jnp.where(is_observation, nobs + 1, nobs)
        if self.return_info:
            info["nobs"] = nobs
        hk.set_state("nobs", nobs)

        if self.min_periods:
            result = jnp.where(nobs >= self.min_periods, mean, jnp.nan)
        else:
            result = mean

        if self.return_info:
            return result, info
        else:
            return result



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define API for Gym environments."""

import logging
from abc import ABCMeta, abstractmethod
from dataclasses import field
from typing import Any, Tuple

from wax.gym.entity import Entity

logger = logging.getLogger(__name__)


class Env(Entity, metaclass=ABCMeta):
    # Set these in ALL subclasses
    action_space: Any = field(init=False, default=None)
    observation_space: Any = field(init=False, default=None)

    @abstractmethod
    def reset(self) -> Any:
        ...

    @abstractmethod
    def step(self, action: Any) -> Tuple[Any, Any, bool, Any]:
        """
        Parameters
        ----------
        action
            action of the agent.

        Returns
        -------
        obs
            Modified observation
        rw
            Reward
        done
            true if the run is over.
        info
            Any data structure containing some general informations
            about the running simulation.

        """

    def close(self):
        ...



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pandas as pd
import pytest

numba = pytest.importorskip("numba")


def test_ewma_numba():
    from wax.numba.ewma_numba import ewma

    x = np.ones((30,), "float64")
    x[0] = np.nan

    x[1] = -1

    x[5:20] = np.nan
    x = x.reshape(-1, 1)
    res, state = ewma(com=10, adjust="linear")(x)
    pd.DataFrame(res).plot()


def check_against_pandas_ewm(x, **ewma_kwargs):
    from wax.numba.ewma_numba import ewma

    res, state = ewma(**ewma_kwargs)(x, state=None)
    res = pd.DataFrame(np.array(res))

    ref_res = pd.DataFrame(x).ewm(**ewma_kwargs).mean()
    pd.testing.assert_frame_equal(res, ref_res, atol=1.0e-6)
    return res


@pytest.mark.parametrize(
    "adjust, ignore_na",
    [(False, False), (False, True), (True, False), (True, True)],  # ,
)
def test_nan_at_beginning(adjust, ignore_na):
    T = 20
    x = np.full((T,), np.nan)
    x[2] = 1
    x[10] = -1

    check_against_pandas_ewm(x, com=10, adjust=adjust, ignore_na=ignore_na)

    # check min_periods option with random variable
    random_state = np.random.RandomState(42)
    x = random_state.normal(size=(5,))
    check_against_pandas_ewm(
        x,
        com=10,
        adjust=adjust,
        ignore_na=ignore_na,
        min_periods=2,
    )

    # check random variable with nans
    x = np.ones((30,), "float64")
    x[0] = np.nan
    x[1] = -1
    x[5:20] = np.nan
    check_against_pandas_ewm(
        x,
        com=10,
        adjust=adjust,
        ignore_na=ignore_na,
    )


def test_init_value():
    from wax.numba.ewma_numba import ewma

    # check random variable with nans
    x = np.ones((30,), "float64")
    x[0] = np.nan
    x[1] = -1
    x[5:20] = np.nan

    x = x.reshape(-1, 1)
    res, state = ewma(com=10, adjust=False, ignore_na=False)(x)

    res_init0, state = ewma(com=10, adjust=False, ignore_na=False, initial_value=0)(x)

    assert res_init0[0] == 0
    assert np.isnan(res[0])
    assert np.linalg.norm(res_init0) < np.linalg.norm(np.nan_to_num(res))


def test_ewma_state():
    from wax.numba.ewma_numba import ewma

    x = np.ones((30,), "float64")
    x[0] = np.nan
    x[1] = -1
    x[5:20] = np.nan

    x = x.reshape(-1, 1)

    ewma_apply = ewma(com=10, adjust="linear", min_periods=5)
    res_full, _ = ewma_apply(x)

    T = 10
    res1, state = ewma_apply(x[:T])
    res2, _ = ewma_apply(x[T:], state)
    res12 = np.concatenate([res1, res2])
    df = pd.concat(
        [pd.DataFrame(res_full), pd.DataFrame(res12)], axis=1, keys=["full", "12"]
    )
    assert len(df)

    assert np.allclose(res_full, res12, equal_nan=True)


@pytest.mark.parametrize("obj_type", ["frame", "series"])
def test_pandas_online(obj_type):
    from wax.numba.ewma_numba import register_wax_numba

    x = np.ones((30,), "float64")
    x[0] = np.nan
    x[1] = -1
    x[5:20] = np.nan

    if obj_type == "frame":
        x = x.reshape(-1, 1)
        X = pd.DataFrame(x)
    else:
        X = pd.Series(x)

    register_wax_numba()
    res_full = X.wax_numba.ewm(com=10).mean()
    res1, state = X.iloc[:10].wax_numba.ewm(com=10, return_state=True).mean()
    res2 = X.iloc[10:].wax_numba.ewm(com=10).mean(state=state)

    res12 = pd.concat([res1, res2])
    if obj_type == "frame":
        pd.testing.assert_frame_equal(res_full, res12)
    else:
        pd.testing.assert_series_equal(res_full, res12)

    assert state.dtypes.tolist() == [np.float64, np.float64, np.int64]



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Forward fill the values."""
import haiku as hk
from jax import numpy as jnp


class Ffill(hk.Module):
    """Ffill current element."""

    def __init__(self, periods=1, name=None):
        super().__init__(name=name)

    def __call__(self, x):
        valid_value = hk.get_state(
            "valid_value",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, jnp.nan, dtype),
        )

        valid_value = jnp.where(jnp.isnan(x), valid_value, x)

        hk.set_state("valid_value", valid_value)
        return valid_value



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute standard deviation  over masked data."""

import haiku as hk
import jax.numpy as jnp
from jax import tree_map

from wax.modules.apply_mask import ApplyMask


class MaskStd(hk.Module):
    """Compute standard deviation  over masked data."""

    def __init__(self, axis=None, assume_centered=False, name=None):
        """Initialize the module.

        Args:
            axis : axis along which to compute the mean
            assume_centered : if True assume mean to be zero when computing
                the standard deviation.
        """
        super().__init__(name=name)
        self.axis = axis
        self.assume_centered = assume_centered

    def __call__(self, mask, input):
        """Compute standard deviation.

        Args:
            mask : bolean mask specifying points over which to compute the mean.
            input : data on which to compute the mean.
        """

        def mask_std(x):
            # mask values
            x = ApplyMask()(mask, x)

            # fill nans
            x = jnp.nan_to_num(x)

            count = mask.sum(axis=self.axis)
            mean = x.sum(axis=self.axis) / count

            if self.assume_centered:
                diff = x
            else:
                diff = x - mean

            diff = ApplyMask()(mask, diff)

            eps = jnp.finfo(x.dtype).resolution ** 2
            var = (diff**2).sum(axis=self.axis) / count
            std = jnp.where(var > 0.0, jnp.sqrt(eps + var), 0.0)
            return std

        return tree_map(mask_std, input)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Optional

import haiku as hk
import jax.numpy as jnp
from haiku.data_structures import partition

from wax.predicate import pass_all_predicate


class UpdateParams(hk.Module):
    """Transform a callable as a function of its trainable parameters.

    The non trainable parameters and state are captured and managed as state
    of the module.
    """

    def __init__(
        self,
        func: Callable,
        params_predicate: Optional[
            Callable[[str, str, jnp.ndarray], bool]
        ] = pass_all_predicate,
        name=None,
    ):
        super().__init__(name=name)
        self.func = func
        self.params_predicate = params_predicate

    def __call__(self, trainable_params, *args, **kwargs):
        init, apply = hk.transform_with_state(self.func)
        rng = hk.next_rng_key() if hk.running_init() else None

        def init_model_non_trainable_params_and_state(shape, dtype):
            """Set state from trainable params and state of the model."""
            params, state = init(rng, *args, **kwargs)

            trainable_params, non_trainable_params = partition(
                self.params_predicate, params
            )
            trainable_params = hk.data_structures.to_mutable_dict(trainable_params)
            return (non_trainable_params, state)

        non_trainable_params, state = hk.get_state(
            "non_trainable_params_and_state",
            [],
            init=init_model_non_trainable_params_and_state,
        )

        params = hk.data_structures.merge(trainable_params, non_trainable_params)

        res, state = apply(params, state, rng, *args, **kwargs)

        hk.set_state("non_trainable_params_and_state", (non_trainable_params, state))
        return res


def get_init_params(
    func,
    *args,
    params_predicate: Optional[
        Callable[[str, str, jnp.ndarray], bool]
    ] = pass_all_predicate,
    **kwargs
):
    init_rng = hk.next_rng_key() if hk.running_init() else None
    init, _ = hk.transform(func)
    params = init(init_rng, *args, **kwargs)

    if params_predicate:
        trainable_params, non_trainable_params = partition(params_predicate, params)
        return trainable_params
    else:
        return params
